import 'dart:async';

import 'package:rxdart/rxdart.dart';

import 'event.dart';
import 'state.dart';
import 'mutation.dart';

/// A dispatcher represents a state machine that can
/// [dispatch] events to trigger a mutation of its
/// [state]
class Dispatcher {
  Dispatcher(State initialState)
      : this._completer = Completer<State>(),
        this._mutating = BehaviorSubject(seedValue: false, sync: true),
        this._state =
            BehaviorSubject<State>(seedValue: initialState, sync: true);

  /// Indicates whether the current [state] is in the process of mutating for a new state.
  bool get isMutating => this._mutating.value;

  /// Raised whenever the state is mutating.
  Stream<bool> get mutating => this._mutating;

  /// Raised each time the [currentState] has mutated to a new state.
  Stream<State> get state => _state.stream;

  /// The current state.
  State get currentState => _state.value;

  BehaviorSubject<State> _state;
  BehaviorSubject<bool> _mutating;
  Completer<State> _completer;
  Map<Type, StateMutator> _mutations = {};
  List<Type> _stop = [];

  /// Register a new [StateMutator] for the given [FromState] state type.
  ///
  /// This will allow to then register mutations for this state type.
  TypedStateMutator<FromState> from<FromState extends State>() {
    final mutator = TypedStateMutator<FromState>();
    _mutations[FromState] = mutator;
    return mutator;
  }

  /// Indicates that [FromState] is a terminal state that should
  /// stop the dispatcher by triggering the [complete] future completion.
  void end<FromState extends State>() => _stop.add(FromState);

  /// Completed when the dispatcher enter in one of its final states
  /// (registered with the [end] function).
  Future complete() => _completer.future;

  /// Indicates whether the current [event] can trigger
  /// a mutation for the [currentState].
  bool canMutate(Event event) {
    final mutator = _mutations[this.currentState.runtimeType];
    return mutator != null && mutator.canMutate(event);
  }

  /// Only [Event] types that can trigger a mutation for the [currentState].
  Iterable<Type> get supportedMutations {
    final mutator = _mutations[this.currentState.runtimeType];
    return (mutator != null) ? mutator.supportedMutations : [];
  }

  /// Indicates whether the dispatcher is in one of its final states (registered
  /// with the [end] function).
  bool get isCompleted =>
      _completer.isCompleted ||
      this._stop.contains(this.currentState?.runtimeType);

  /// Dispatch an event to trigger a mutation of the [currentState].
  ///
  /// If an event is raised while the dispatcher is already [mutating], it is
  /// ignored.
  Future dispatch(Event event) async {
    if (!this.isMutating && !this.isCompleted) {
      final mutator = _mutations[this.currentState.runtimeType];
      if (mutator.canMutate(event)) {
        try {
          this._mutating.add(true);
          final newState = await mutator.mutate(this.currentState, event);
          this._mutating.add(false);
          this._state.add(newState);
          if (this.isCompleted) {
            this._completer.complete(newState);
          }
        } catch (e) {
          this._completer.completeError(e);
          throw e;
        }
      } else {
        this._state.add(this.currentState);
      }
    }
  }

  /// Dispose the instance by closing all subscriptions.
  void dispose() {
    this._mutating.close();
    this._state.close();
  }
}
