import 'dart:async';

import 'package:mutation/mutation.dart';

/// A mutation of the current [state] to a new [State] from the given [event].
typedef FutureOr<State> Mutation(State state, Event event);

/// A set of mutations that are reacting to supported events.
abstract class StateMutator {
  Map<Type, Mutation> _mutations = {};

  /// Indicates whether the current [event] can trigger
  /// a mutation.
  bool canMutate(Event e) => _mutations.containsKey(e?.runtimeType);

  /// Only [Event] types that can trigger a mutation.
  Iterable<Type> get supportedMutations => this._mutations.keys;

  /// Mutate the given [state] to a new [State] from the given [event].
  ///
  /// The [event] must be a supported event type ([canMutate] should return [true] for it).
  FutureOr<State> mutate(State state, Event event) {
    final eventType = event?.runtimeType;
    final mutation = _mutations[eventType];
    if (mutation == null) {
      throw Exception("not supported event type $eventType");
    }
    return mutation(state, event);
  }
}

/// A mutator that can only mutate states of type [S].
class TypedStateMutator<S extends State> extends StateMutator {
  /// Register a [mutation] when an event of type [E] is given to [mutate] function.
  TypedStateMutator<S> on<E extends Event>(
      FutureOr<State> mutation(S state, E event)) {
    _mutations[E] = (s, e) => mutation(s as S, e as E);
    return this;
  }
}
